#!/usr/bin/env node

const fs = require('fs'),
    _ = require('underscore'),
    mysql = require('mysql'),
    moment = require('moment');

let fileUtils = {
    walkFile(pathResolve, mime) {
        let files = fs.readdirSync(pathResolve),
            fileList = {};
        for (let [i, item] of files.entries()) {
            let itemArr = item.split('\.'),
                itemMime = (itemArr.length > 1) ? itemArr[itemArr.length - 1] : 'undefined',
                keyName = item + '';
            if (mime === itemMime) {
                fileList[item] = pathResolve + item;
            }
        }
        return fileList;
    },

    getFileData(fileName, path) {
        return fs.readFileSync(path, 'binary');
    }
}

let sqlUtils = {
    cfg: null,
    pool: null,
    sqlContentList: [],
    lastVersion: 0,
    init(params) {
        this.cfg = {
            sqlDir: params[0],
            dbSvr: params[1],
            dbName: params[2],
            rootName: params[3],
            rootPass: params[4],
            userName: params[5],
            userPass: params[6]
        };
        console.log(this.cfg);
        this.pool = mysql.createPool({
            host: this.cfg.dbSvr,
            user: this.cfg.rootName,
            password: this.cfg.rootPass
        });
    },

    async query(sql, values) {
        var _self = this;
        return new Promise(function(resolve, reject) {
            _self.pool.getConnection(function(err, connection) {
                if (err) {
                    reject(err);
                } else {
                    connection.query(sql, values, function(err, rows) {
                        if (err) {
                            console.log(err);
                            reject(err);
                        } else {
                            resolve(rows);
                        }
                        connection.release();
                    })
                }
            })
        })
    },

    async checkDataBaseExist() {
        // 检查数据库是否存在
        var checkSql = "select count(*) as tableLen from INFORMATION_SCHEMA.TABLES where TABLE_SCHEMA=?";
        console.log("\n\ndbName:", this.cfg.dbName);
        let result = await this.query(checkSql, [this.cfg.dbName]);
        return result[0].tableLen !== 0;
    },

    async createDatabase() {
        // 创建数据库
        // let createSql = "create database " + this.cfg.dbName + " CHARACTER SET utf8 COLLATE utf8_general_ci",
        let createSql = "create database " + this.cfg.dbName + " CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci",
            result = await this.query(createSql);
        return result.affectedRows !== 0;
    },

    async addDbUser() {
        // 添加数据库用户
        // var addUserSql = "grant all privileges on " + this.cfg.dbName + ".* to " + this.cfg.userName + " identified by '" + this.cfg.userPass + "'",
        await this.query("create user " + this.cfg.userName + "@'%' identified by '" + this.cfg.userPass + "';")
        await this.query('flush privileges;')
        await this.query("grant all privileges on " + this.cfg.dbName + ".* to " + this.cfg.userName + "@'%' with grant option;")
        return true;
    },

    async createSqlVersionTable() {
        // 创建版本管理表
        var createSqlVersionTableSql = "create table " + this.cfg.dbName + ".sql_version(name varchar(512))",
            result = await this.query(createSqlVersionTableSql);
        return true;
    },

    async useDb() {
        // 选中数据库
        await this.query("use " + this.cfg.dbName);
    },

    async getLastVersion() {
        // 获取上一次版本
        var getVersionSql = "select name from " + this.cfg.dbName + ".sql_version";
        var result = await this.query(getVersionSql);
        return (result && result.length !== 0) ? result[0].name : 0
    },

    async updateVersion() {
        // 更新版本号
        var deleteVersionSql = "delete from " + this.cfg.dbName + ".sql_version";
        var deleteResult = await this.query(deleteVersionSql);
        var updateVersionSql = "insert into " + this.cfg.dbName + ".sql_version values('" + this.lastVersion + "')";
        var updateResult = await this.query(updateVersionSql);
        return true
    },

    getSqlFileMap() {
        // 获取 script 目录下的 所有 .sql 文件, 返回 {fileName: filtPath}
        var basePath = process.cwd(); // __dirname \ __filename
        basePath = basePath + '/' + this.cfg.sqlDir;
        return fileUtils.walkFile(basePath, 'sql');
    },

    getSqlContentMap() {
        var _self = this;
        // 封装所有sql文件脚本内容
        var sqlMap = this.getSqlFileMap();
        console.log("sqlMap:", sqlMap);
        for (var versionFile in sqlMap) {
            var version = versionFile.split('.')[0];
            console.log("version:", version);
            if ((version === 'create' && this.lastVersion === 0) || (version !== 'create' && version > this.lastVersion && this.lastVersion != 0)) {
                console.log("new version file");
                var content = fileUtils.getFileData(versionFile, sqlMap[versionFile]);
                var contentList = content.split("\n");

                var tempSql = [];
                var index = 0;
                _.each(contentList, function(sqlItem) {
                    // 过滤掉注释
                    if (sqlItem.indexOf("--") === -1) {
                        // 通过识别 ";" 来判断一条语句是否结束
                        if (sqlItem.indexOf(";") === -1) {
                            tempSql.push(sqlItem);
                        } else {
                            tempSql.push(sqlItem);
                            _self.sqlContentList.push(tempSql.join(""));
                            tempSql = [];
                            index += 1;
                        }
                    }
                })
            }
        }
    },

    async createAllTables() {
        // 判断数据库是否存在，不存在的话创建数据库
        let DataBaseExist = await this.checkDataBaseExist();
        if (!DataBaseExist) {
            console.log("数据库还未创建");
            console.log("============= 开始创建数据库 =============");
            let createResult = await this.createDatabase();
            if (createResult) {
                let addDbUserResult = await this.addDbUser();
                let createVersionResult = await this.createSqlVersionTable();
                await this.useDb();
            }
            console.log("============= 创建数据库完成 =============")
        } else {
            await this.useDb();
            this.lastVersion = await this.getLastVersion();
            console.log("last version:", this.lastVersion);
        }

        // 如果不是创建数据库，那么从数据库获取最后一次的版本，然后便利当前 sql 文件目录是否有比这个值大的版本

        // 执行 sql
        // 获取所有sql脚本内容
        this.getSqlContentMap();
        var sqlContentList = this.sqlContentList;
        // console.log("sqlContentList:", sqlContentList)
        if (sqlContentList.length !== 0) {
            // 执行建表sql脚本
            for (var i = 0; i < sqlContentList.length; i++) {
                let result = await this.query(sqlContentList[i]);
                if (result.serverStatus * 1 === 2) {
                    console.log('success');
                } else {
                    console.log(sqlContentList[i]);
                    console.log('fail');
                }
            }
            this.lastVersion = moment(new Date()).format("YYYYMMDDHHmmssSSS");
            console.log(this.lastVersion);
            await this.updateVersion();
            console.log('sql脚本执行结束!\n\n');
        } else {
            console.log('没有数据库需要更新\n\n');
        }
        process.exit();
    }
};

// 命令行参数： nodejs node_file param1...

// 自动实现数据版本管理:
// 没有管理表的话,创建版本表，用 create.sql 来创建
// 有表的话,读取最后一次版本,并读取指定目录下的大于这个版本的 sql 文件
// sql 的话,从文件中读取,然后解析成一条条语句的数组,然后一条条执行

// 参数列表：
// sql文件所在目录、数据库地址、数据库名、root 账号密码、数据库用户账号密码(7个参数)

async function main(params) {
    await sqlUtils.init(params);
    await sqlUtils.createAllTables();
}

if (process.argv.length === 9) {
    main(process.argv.slice(2, 9));
} else {
    console.error("param error, example: node sql_version.js sql_path dbSvr dbName rootName rootPass userName userPass");
}
